--==================================================================
--Generic LFSR Operation and LFSR Clock Divider
--This package and components provide generic LFSR Maximal-Length operation for any 
--  length from 3 to 168. Except 37! If you need a register width of 37 you are on your own,
--  since we were too lazy to add a special case for 6 taps. The package contains all the 
--  constants and functions needed for usage of two components included in the file:
--  A generic lfsr, and a generic lfsr clock divider that decodes the lfsr state to pulse
--  the output clock like a modulo N clock divider. From the generic division parameter N, 
--  the lfsr width and the decode value are automatically computed as compile time constants.
--  For reasons discussed below, the lfsr clock division for divisors > 10,000 may not work. 
--The "lfsr_div" entity will select the minimum sized LFSR to divide the input 
--  clock with a decoded, modulo-N LFSR operation restricted to N less than 10,000. The  
--  reason for the restriction is to avoid the notorious ISE exception error occuring   
--  with large for-loops, that arises while computing a decode state. We know ISE 9.2
--  works up to N=10_000 for our environment, but may trigger the exception for yours.
--  If so, you can just instantiate two cascaded lfsr_div components to complete the division.
--  If the DIVISOR is exactly 2**N-1, then this problem never arises, since the
--  decode state is just the final rollover state, which is easy to find for any length. 
--The generic "lfsr" entity has no such problem. It operates on the clock rising edge for any generically supplied
--  shift register WIDTH. It ports out the register state and the PRN output 
--  bit. These can be externally latched if desired on the falling edge of the LFSR clock,   
--  since the shifts occur on the rising edge.
--Ref. XILINX Application Note XAPP 052, Peter Alfke, July 7, 1996.
------------------------------------- 
--Copywrite Johns Hopkins University
---ECE Department - Josh Blum, Robert E. Jenkins
-------------------------------------
--This package and components may be freely used as long as this header is retained.
--==================================================================
------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use IEEE.math_real.all;
------------------------------
---**************************************
PACKAGE lfsr_package is
---**************************************

--LFSR Feedback taps subscripts are specified in an array that is 4 integers wide.
--The tap-subscript entry corresponding to the requested LFSR size is selected at compile time, 
--  and then 0 entries are ignored in the xnor implementation.
constant DEFAULT_DIV: integer:= 5451873;	--default generic divisor in lfsr_div, used for testing.  
constant TAP_SIZE: integer := 168; 
constant MIN_TAP_SIZE: integer := 3; --taps start with size 3 LFSR

---Array of taps subscripts-----------------
type SUBS_TYPE is array(3 downto 0) of integer;		--type definitions for arrays
type SUB_ARRAY_TYPE is array(3 to TAP_SIZE) of SUBS_TYPE;
constant TAP_SUBS_ARRAY: SUB_ARRAY_TYPE:= (
--(0,0,2,1),
(0,0,3,2),	--size 3
(0,0,4,3),
(0,0,5,3),
(0,0,6,5),
(0,0,7,6),
(8,6,5,4),
(0,0,9,5),
(0,0,10,7),	--size 10
(0,0,11,9),
(12,6,4,1),
(13,4,3,1),
(14,5,3,1),
(0,0,15,14),
(16,15,13,4),
(0,0,17,14),
(0,0,18,11),
(19,6,2,1),
(0,0,20,17), --size 20
(0,0,21,19),
(0,0,22,21),
(0,0,23,18),
(24,23,22,17),
(0,0,25,22),
(26,6,2,1),
(27,5,2,1),
(0,0,28,25),
(0,0,29,27),
(30,6,4,1),	--size 30
(0,0,31,28),
(32,22,2,1),
(0,0,33,20),
(34,27,2,1),
(0,0,35,33),
(0,0,36,25),
(4,3,2,1),	--this is a dummy entry. size 37 not allowed
(38,6,5,1),
(0,0,39,35),
(40,38,21,19),	--size 40
(0,0,41,38),
(42,41,20,19),
(43,42,38,37),
(44,43,18,17),
(45,44,42,41),
(46,45,26,25),
(0,0,47,42),
(48,47,21,20),
(0,0,49,40),
(50,49,24,23),	--50
(51,50,36,35),
(0,0,52,49),
(53,52,38,37),
(54,53,18,17),
(0,0,55,31),
(56,55,35,34),
(0,0,57,50),
(0,0,58,39),
(59,58,38,37),
(0,0,60,59),	--60
(61,60,46,45),
(62,61,6,5),
(0,0,63,62),
(64,63,61,60),
(0,0,65,47),
(66,65,57,56),
(67,66,58,57),
(0,0,68,59),
(69,67,42,40),
(70,69,55,54),	--70
(0,0,71,65),
(72,66,25,19),
(0,0,73,48),
(74,73,59,58),
(75,74,65,64),
(76,75,41,40),
(77,76,47,46),
(78,77,59,58),
(0,0,79,70),
(80,79,43,42),	--80
(0,0,81,77),
(82,79,47,44),
(83,82,38,37),
(0,0,84,71),
(85,84,58,57),
(86,85,74,73),
(0,0,87,74),
(88,87,17,16),
(0,0,89,51),
(90,89,72,71),	--90
(91,90,8,7),
(92,91,80,79),
(0,0,93,91),
(0,0,94,73),
(0,0,95,84),
(96,94,49,47),
(0,0,97,91),
(0,0,98,87),
(99,97,54,52),
(0,0,100,63),	--100
(101,100,95,94),
(102,101,36,35),
(0,0,103,94),
(104,103,94,93),
(0,0,105,89),
(0,0,106,91),
(107,105,44,42),
(0,0,108,77),
(109,108,103,102),
(110,109,98,97),	--110
(0,0,111,101),
(112,110,69,67),
(0,0,113,104),
(114,113,33,32),
(115,114,101,100),
(116,115,46,45),
(117,115,99,97),
(0,0,118,85),
(0,0,119,111),		--120
(120,113,9,2),
(0,0,121,103),
(122,121,63,62),
(0,0,123,121),
(0,0,124,87),
(125,124,18,17),
(126,125,90,89),
(0,0,127,126),
(128,126,101,99),
(0,0,129,124),		--130
(0,0,130,127),
(131,130,84,83),
(0,0,132,103),
(133,132,82,81),
(0,0,134,77),
(0,0,135,124),
(136,135,11,10),
(0,0,137,116),
(138,137,131,130),
(139,136,134,131),	--140
(0,0,140,111),
(141,140,110,109),
(0,0,142,121),
(143,142,123,122),
(144,143,75,74),
(0,0,145,93),
(146,145,87,86),
(147,146,110,109),
(0,0,148,121),		--150
(149,148,40,39),
(0,0,150,97),
(0,0,151,148),
(152,151,87,86),
(0,0,153,152),
(154,152,27,25),
(155,154,124,123),
(156,155,41,40),
(157,156,131,130),	--160
(158,157,132,131),
(0,0,159,128),
(160,159,142,141),
(0,0,161,143),
(162,161,75,74),
(163,162,104,103),
(164,163,151,150),
(165,164,135,134),
(166,165,128,127),	--170
(0,0,167,161),
(168,166,153,151) ); 

----function declarations------------------------
--shift an lfsr back one step
function BACK_LFSR(reg: in std_logic_vector; tap_subs: in SUBS_TYPE) return std_logic_vector;
--find the lfsr state after N steps starting from (others=>'0')
function find_lfsr_decode(steps: in integer; tap_subs: in SUBS_TYPE; width:in integer) return std_logic_vector;  
--find the minimum lfsr width to do a mod N clock division
function find_lfsr_size(divisor: in integer) return integer;
--shift an lfsr forward one step
function SHFT_LFSR(reg: in std_logic_vector; tap_subs: in SUBS_TYPE) return std_logic_vector;
-------------------------------------------------

component lfsr_div 	--LFSR clock divider with auto decode and width detect
  generic ( DIVISOR: integer := DEFAULT_DIV );
  port (
	rst: 		in std_logic;  --active high asyncronous reset
	clk_in: 	in std_logic;  --input clock
	clk_out: out std_logic  --output clock
	);  
end component;
-----------------------------------------------
component lfsr		--clocked LFSR of generic size
  generic ( WIDTH: integer := 4 );
  port (
	rst: 		in std_logic;  	--active high asyncronous reset
	enb: 		in std_logic; 		--active high asyncronous enable
	clk_in: 	in std_logic;  	--input clock
	state:	out std_logic_vector(WIDTH downto 1);	--lfsr state
	bit_out: out std_logic  --output PRN bit
	);  
end component;
-------------------------------------------------------
component lfsr_clk is 	--LFSR clock divider where width and decode state is supplied
generic (WIDTH: integer:= 4;
			DECODE: integer:= 14	 --decode state for N shifts
			);	--
  port (
	rst: in std_logic;  		--active high asyncronous reset
	clk_in: 	in std_logic;  --input clock
	state: 	out std_logic_vector(WIDTH downto 1); --for simulation debug
	clk_out: out std_logic  --output clock
	);  
end component;
-----------------------------------------------
---*******************************************
END PACKAGE lfsr_package; 
---*******************************************
  
---*******************************************
PACKAGE BODY lfsr_package  is
---*******************************************  
--Function to move an lfsr backward one step using TAPS subscripts
  function BACK_LFSR(reg: in std_logic_vector; tap_subs: in SUBS_TYPE) return std_logic_vector is
    variable old_msb: std_logic;
    begin
	 old_msb:= reg(1+tap_subs(0)) xor not reg(1);
	 if tap_subs(3) /= 0 then 
	   old_msb:= old_msb xnor reg(1+tap_subs(3)) xnor reg(1+tap_subs(2)); 
	 end if;
    return old_msb & reg(reg'HIGH downto 2); --shift in the old MSB
  end function;
----------------------------------------------
--Function to move an lfsr forward one step using TAPS subscripts
  function SHFT_LFSR(reg: in std_logic_vector; tap_subs: in SUBS_TYPE) return std_logic_vector is
    variable xnorred: std_logic;
    begin
	 xnorred:= reg(tap_subs(0)) xnor reg(tap_subs(1)); --these two taps are always present.
	 if tap_subs(3) /= 0 then 
	   xnorred:= xnorred xnor reg(tap_subs(3)) xnor reg(tap_subs(2));
	 end if;
    return reg(reg'HIGH - 1 downto reg'LOW) & xnorred; --shift out the old MSB
  end function;
-----------------------------------------------
--Compile-time function to get the lfsr decode state after N steps
function find_lfsr_decode(steps: in integer; tap_subs: in SUBS_TYPE; width:in integer) return std_logic_vector is
	 variable reg: std_logic_vector(width downto 1);
    begin
    reg := (others => '0');
    for i in 1 to steps-1 loop
      reg := SHFT_LFSR(reg, tap_subs);
    end loop;
    return reg;
  end function;
-----------------------------------------------
--Function to get the minimum lfsr size n, such that 2^n - 1 > divisor
function find_lfsr_size(divisor: in integer) return integer is
variable size: integer; 
  begin
  size:= integer(ceil( log(real(divisor))/log(real(2) ) ) );  --n>log2(divisor)
  if size < 3 then 
		return 3;
  elsif size = 37 then
		return 38;
  else 
		return size;  
  end if;
end function;
------------------------------------------------

----*********************************************
END PACKAGE BODY lfsr_package;
----*********************************************

--======================================================================
-- Divide a clock by N (DIVISOR) using a Maximal length LFSR as a modulo N counter.
-- DIVISOR is specified as a generic parameter greater than 1.
-- Chooses smallest maximal-length cycle LFSR and its decode condition for N steps.
-- The algorithm picks the LFSR taps from the taps array, size 3 to 168.
--Because of the large FOR loops, this only works for divisors less than about 10_000. 
-------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;
use WORK.lfsr_package.all;

-------------------------------------------------------
entity lfsr_div is --automated clock divider
  generic ( DIVISOR: integer := DEFAULT_DIV );	--
  port (
	rst: in std_logic;  	--active high asyncronous reset
	clk_in: in std_logic;  	--input clock
	clk_out: out std_logic  --output clock
	);  
end entity lfsr_div;

--------------------------------------------------------------------
architecture ARCH of lfsr_div is
-------------------------------------------------
--LFSR functions select minimum width and tap subscripts for that width.
constant WIDTH: integer := find_lfsr_size(DIVISOR);	--minimum width for DIVISOR
 --get taps subscripts	for given width					
constant TAP_SUBS1: SUBS_TYPE:= TAP_SUBS_ARRAY(WIDTH);	
	 
--the shift registers and associated signals
signal lfsr1: std_logic_vector (WIDTH downto 1);
signal decode: std_logic_vector(WIDTH downto 1);	--decode state for DIVISOR steps
constant ZERO: std_logic_vector(WIDTH downto 1):=(others=>'0'); --just a way to ref a 0 state
signal xnorred1: std_logic;
  
  begin

---================================================================
-- If synthesizable, generate the code for a single lfsr to divide the clock. 
  --first get the lfsr decode state for the case that applies
  CASE1: if (DIVISOR=2**WIDTH-1) generate
		decode<= BACK_LFSR(ZERO, TAP_SUBS1);
  end generate CASE1;
  CASE2: if (DIVISOR /= 2**WIDTH -1) generate
		decode<= find_lfsr_decode(DIVISOR, TAP_SUBS1, WIDTH);
  end generate CASE2;
  --next run a modulo N clock divider by pulsing at the lfsr decode state. 
	SHIFT: process (rst, clk_in) is begin
      if rst = '1' then				--asyncronous reset
        lfsr1 <= (others => '0');	--reset to all zeros
        clk_out <= '0';
      elsif rising_edge(clk_in) then
			 if lfsr1 = decode then
            --reset the lfsr and set clk to high
            lfsr1 <= (others => '0');
            clk_out <= '1';
          else
            --shift the lfsr. and set clk low
				lfsr1 <= lfsr1(lfsr1'HIGH - 1 downto 1) & xnorred1;
				clk_out <= '0';
          end if;
      end if;
    end process SHIFT;
--------------------------------------------
---Generate xnor code that applies, based on the tap subscripts for WIDTH
--------------------------------------------
TWO: if TAP_SUBS1(2)= 0 generate	--There can only be 2 or 4 taps (we dont use W=37 to avoid 6)
	xnorred1<= lfsr1(tap_subs1(0)) xnor lfsr1(tap_subs1(1));
end generate TWO;

FOUR: if (TAP_SUBS1(3) /= 0) generate
	xnorred1<= lfsr1(tap_subs1(0)) xnor lfsr1(tap_subs1(1)) xnor lfsr1(tap_subs1(2)) xnor lfsr1(tap_subs1(3));
end generate FOUR;
------------------------------

end architecture ARCH;

--========================================================================
--Operate an LFSR of generic width that ranges from 3 thru 168.
--LFSR functions from lfsr package generates taps vector for the specified width.
--  The algorithm picks the lfsr taps from an array of taps, of width 168.
--The register state and the output PRN bit are output port signals.
-----------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.math_real.all;
use WORK.lfsr_package.all;

-----------------------------------------------
entity lfsr is 
  generic (WIDTH: integer := 4 );	--generic lfsr size, allowed range 3-168 (except for 37)
  port (
	rst: 		in std_logic;  			--active high asyncronous reset
	enb: 		in std_logic; 				--active high asyncronous enable
	clk_in: 	in std_logic;  			--input clock
	state:	out std_logic_vector(WIDTH downto 1);	--lfsr state 
	bit_out: out std_logic  			--output bit from shift
	);  
end entity lfsr;

-------------------------------------------
architecture ARCH of lfsr is
-------------------------------------------
--LFSR function from lfsr package selects taps vector for the specified width.						
  constant TAP_SUBS: SUBS_TYPE:= TAP_SUBS_ARRAY(WIDTH);							
  --the shift register
  signal lfsr: std_logic_vector (WIDTH downto 1);
  signal xnorred, xnorred1: std_logic;
  
  begin
 state<= lfsr;	--output the register state 
 SHIFT: process (rst, enb, clk_in, lfsr, xnorred) is begin
    if rst = '1' then		--asyncronous reset
		lfsr <= (others => '0');	--reset LFSR to all zeros
    elsif enb = '1' then 	--async enable
      if rising_edge(clk_in) then   
          lfsr <= lfsr(WIDTH-1 downto 1) & xnorred;
			 bit_out<= lfsr(WIDTH);
      end if;
    end if;
 end process SHIFT;
 
 ------------------------------------------
---Generate xnor code based on the tap subscripts for this width
-------------------------------------------
--There can only be 2 or 4 taps (excepting width 37) 
	xnorred1<= lfsr(tap_subs(0)) xnor lfsr(tap_subs(1)); --1st 2 always apply

FOUR: if TAP_SUBS(3) /= 0 generate
	xnorred<= xnorred1 xnor lfsr(tap_subs(2)) xnor lfsr(tap_subs(3));
end generate FOUR;

--------------------------------------------
end architecture ARCH;
--------------------------------------------

--======Divide clock with a Supplied Decode state==========
--Divide a clock with an input decode state using a Maximal length LFSR as a mod N counter.
--WIDTH is specified as a generic parameter greater than 2.
--Decode value is specific as a generic integer, if zero then rolls over after 2**N - 1.
--The package picks the LFSR taps from the taps array, size 3 to 168. 
---------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
--use ieee.math_real.all;
use WORK.lfsr_package.all;
---------------------------------------------
entity lfsr_clk is 
generic (WIDTH: integer:= 4;
			DECODE: integer:= 14  --The lfsr state for the desired decode.
			);
port (
	rst: 		in std_logic;  	--active high asyncronous reset
	clk_in: 	in std_logic;  	--input clock
	state:	out std_logic_vector(WIDTH downto 1); --for simulation. else may be left open.
	clk_out: out std_logic  --output clock
	);  
end entity lfsr_clk;

---------------------------------------------
architecture ARCH of lfsr_clk is
---------------------------------------------
--LFSR function TAP_SUBS_ARRAY(WIDTH) selects 4 tap subscripts for any given width.					
constant TAP_SUBS1: SUBS_TYPE:= TAP_SUBS_ARRAY(WIDTH);		 
--the shift registers and associated signals
signal lfsr: std_logic_vector (WIDTH downto 1);
constant ZERO: std_logic_vector(WIDTH downto 1):=(others=>'0'); --just a way to ref a 0 state
signal xnorred, xnorred1: std_logic;
--the decode state for N steps
constant MOD_ST: std_logic_vector(WIDTH downto 1):=std_logic_vector(TO_UNSIGNED(DECODE,WIDTH));
--attribute s: string;
--attribute s of lfsr:signal is "yes";
begin
-------------------------------------
--Run a modulo N clock divider by pulsing at the lfsr decode state.
state<= lfsr;	--state output for simulation debug 
SHIFT: process (rst, clk_in) is begin
   if rst = '1' then				--asyncronous reset
      lfsr <= (others => '0');	--reset to all zeros
      clk_out <= '0';
   elsif rising_edge(clk_in) then
		if lfsr = MOD_ST then
			clk_out <= '1'; --reset the lfsr and set clk to high
			lfsr <= (others => '0');
		else
			--default is shift the lfsr. and hold clk low
			lfsr <= lfsr(lfsr'HIGH - 1 downto 1) & xnorred;
			clk_out <= '0';		
		end if;
   end if;
end process SHIFT;
-----------------------------------
---Generate applicable xnor code, based on the tap subscripts for WIDTH
-----------------------------------
--There can only be 2 or 4 taps (we never use WIDTH of 37 to avoid 6 taps)
xnorred1<= lfsr(tap_subs1(0)) xnor lfsr(tap_subs1(1));

FOUR: if (TAP_SUBS1(3) /= 0) generate
	xnorred<= xnorred1 xnor lfsr(tap_subs1(2)) xnor lfsr(tap_subs1(3));
end generate FOUR;
------------------------------

end architecture ARCH;

--========================================================================

